# -*- encoding: utf-8 -*-
""" asset archive management
"""
import time
from xml.sax import parseString, ContentHandler
from xml.sax.saxutils import escape

from utils import timetag
from config import FRS_ARCHIVED_FOLDER_NAME


class ArchiveSupportMixin:

    def archivedpath(self, path, *args):
        return self.frspath(path, FRS_ARCHIVED_FOLDER_NAME, *args)

    def archive(self, path, id):
        archivePath = self.archivedpath(path)
        if not self.exists(archivePath):
            self.makedirs(archivePath)

        dstpath = self.joinpath(archivePath, id)
        self.copyAsset(path, dstpath, copycache=0)

    def listArchives(self, path):
        archivePath = self.archivedpath(path)
        if self.exists(archivePath):
            archives = self.listdir(archivePath)
        else:
            archives = []
        return archives

    def copyArchive(self, path, archiveName, dstPath):
        assetpath = self.archivedpath(path, archiveName)
        return self.copyAsset(assetpath, dstPath, copycache=0)

    def removeArchive(self, path, archiveName):
        self.removeAsset( self.archivedpath(path, archiveName) )




# -*- encoding:utf-8 -*-
""" vfs = virtual file system; a virtual posix like file system
"""

import os, shutil
import posixpath
from types import UnicodeType
from ConfigParser import SafeConfigParser
from StringIO import StringIO


from utils import ucopy2, ucopytree, umove

from archive import ArchiveSupportMixin
from recycle import RecycleMixin
from cache import CacheMixin
from config import FRS_ARCHIVED_FOLDER_NAME

def loadFRSFromConfig(config):
    """ load frs from config file """
    cp = SafeConfigParser()
    cp.readfp(StringIO(config))

    frs = FRS()
    cache_path = cp.get('cache', 'path')
    frs.setCacheRoot(unicode(cache_path))
    
    roots = cp.items('root')
    for name,path in roots:
        path = os.path.normpath(path)
        if not os.path.exists(path):
            os.makedirs(path)
        frs.mount(unicode(name), unicode(path))

    roots = cp.items('site')
    for site_path, vpath in roots:
        frs.mapSitepath2Vpath( unicode(site_path), unicode(vpath) )
    return frs

class FRS(ArchiveSupportMixin, RecycleMixin, CacheMixin):

    def __init__(self, cache_root='/tmp', dotfrs='.frs', version="json"):
        self._top_paths = {}
        self.dotfrs = dotfrs
        self.cache_root = cache_root
        self.sitepaths2vpaths = []
        self.version = version

    def mount(self, name, path):
        """ XXX only support mount top dirs only now

        mount filesystem path to vfs
        """
        #if not os.path.exists(path):
        #    raise OSError('no mount path: '+ path)
        if name.startswith('/'):
            name = name[1:]
        self._top_paths[name] = path

    def setCacheRoot(self, path):
        """ where to push caches """
        self.cache_root = path
        self.mount('cache', path)

    def mapSitepath2Vpath(self, site_path, vpath):
        """ map vpath to site path """
        self.sitepaths2vpaths.append( (site_path, vpath) )

    def sitepath2Vpath(self, site_path):
        if site_path[-1] != '/':
            site_path += '/'

        # 历史版本，直接找到对应的历史版本文件夹
        if '/++versions++/' in site_path:
            site_path, version = site_path.split('/++versions++/')
            site_path = site_path.split('/')
            site_path.insert(-1, self.dotfrs)
            _, ext = os.path.splitext(site_path[-1])
            site_path.append(FRS_ARCHIVED_FOLDER_NAME)
            site_path.append(version + ext)
            site_path = '/'.join(site_path)

        for _spath, _vpath in self.sitepaths2vpaths:
            if _spath[-1] != '/':
                _spath += '/'

            if site_path.startswith(_spath):
                if _vpath[-1] != '/':
                    _vpath += '/'

                result = _vpath + site_path[len(_spath):]
                if result[-1] == '/':
                    return result[:-1]
                return result
        raise ValueError('can not find a frs path for site path %s' % site_path)

    def vpath(self, ospath):
        """ transform ospath to vpath """
        for root, path in self._top_paths.items():
            if ospath.startswith(path + '/'):
                return '/%s%s' % (root, ospath[len(path):])

    def ospath(self, vPath):
        """ transform to a real os path """
        if not vPath.startswith('/'):
            raise OSError(vPath)
        parts = vPath.split('/')
        try:
            toppath = self._top_paths[parts[1]]
        except KeyError:
            if parts[1] == self.dotfrs:
                try:
                    toppath = self._top_paths[parts[2]]
                except:
                    raise OSError(vPath)
                basename = os.path.basename(toppath)
                basedir = os.path.dirname(toppath)
                return os.path.join(basedir, self.dotfrs, basename, *parts[3:])
            raise OSError(vPath)
        return os.path.join(toppath, *parts[2:])

    def frspath(self, vpath, *frs_subpaths):
        """ It is another kind of joinpath, which returns path in the .frs folder.
        """
        return self.joinpath(self.dirname(vpath), \
                  self.dotfrs, self.basename(vpath), *frs_subpaths)

    def exists(self, vPath):
        try:
            path = self.ospath(vPath)
        except OSError:
            return False
        return os.path.exists(path)

    def joinpath(self, *arg):
        return posixpath.join(*arg)

    def basename(self, path):
        return posixpath.basename(path)

    def splitext(self, name):
        return os.path.splitext(name)

    def stat(self, vPath):
        return os.stat(self.ospath(vPath))

    def dirname(self, path):
        return posixpath.dirname(path)

    def ismount(self, vPath):
        """ return if vPath is a mount folder
        """
        return vPath[1:] in self.listdir('/')

    def isdir(self, vPath):
        return os.path.isdir(self.ospath(vPath))

    def isfile(self, vPath):
        return os.path.isfile(self.ospath(vPath))

    def atime(self, vPath):
        return os.path.getatime( self.ospath(vPath) )

    def mtime(self, vPath):
        return os.path.getmtime( self.ospath(vPath) )

    def ctime(self, vPath):
        return os.path.getctime( self.ospath(vPath) )

    def getsize(self, vPath):
        return os.path.getsize( self.ospath(vPath) )

    def listdir(self, vPath, pattern=None):
        if vPath == '/':
             return self._top_paths.keys()
        names = os.listdir(self.ospath(vPath))
        if pattern is not None:
            names = fnmatch.filter(names, pattern)

        names = [name for name in names if not name.startswith(self.dotfrs) ]
        return names

    def dirs(self, vPath, pattern=None):
        names = [ name for name in self.listdir(vPath, pattern)\
	                  if self.isdir(self.joinpath(vPath, name))]
        return names

    def files(self, vPath, pattern=None):
        names = [ name for name in self.listdir(vPath, pattern)\
	                  if self.isfile(self.joinpath(vPath, name))]
        return names

    def open(self, vPath, mode='r'):
        return file(self.ospath(vPath), mode)

    def move(self, vPath1, vPath2):
        # can't remove mount folder
        if self.ismount(vPath1):
            raise Exception("can't remove mount folder %s" % vPath1)
        if self.ismount(vPath2):
            raise Exception("can't move to mount folder %s" % vPath2)

        umove(self.ospath(vPath1), self.ospath(vPath2) )
        #notify(AssetMoved(self, vPath1, vPath2))

    def mkdir(self, vPath, mode=0777):
        os.mkdir(self.ospath(vPath), mode)

    def makedirs(self, vPath, mode=0777):
        os.makedirs(self.ospath(vPath), mode)

    def getNewName(self, path, name):
        while self.exists(self.joinpath(path, name)):
            name = 'copy_of_' + name
        return name

    def remove(self, vPath):
        """ remove a file path"""
        os.remove(self.ospath(vPath) )
        #notify(AssetRemoved(self, vpath))

    def copyfile(self, vSrc, vDst):
        shutil.copyfile(self.ospath(vSrc), self.ospath(vDst))

    def copytree(self, vSrc, vDst):
        # copy2 don't work well with encoding
        # in fact it is os.utime don't work well
        ossrc = self.ospath(vSrc)
        osdst = self.ospath(vDst)
        ucopytree(ossrc, osdst, symlinks=False)

    def rmtree(self, vPath, ignore_errors=False, onerror=None):
        # can't remove mount folder
        if self.ismount(vPath):
            raise Exception("can't remove mount folder %s" % vPath)

        shutil.rmtree(self.ospath(vPath), ignore_errors, onerror)

    def touch(self, vpath):
        fd = os.open(self.ospath(vpath), os.O_WRONLY | os.O_CREAT, 0666)
        os.close(fd)

    def walk(self, top, topdown=True, onerror=None):
        if top == '/':
            mount_dirs = self._top_paths.keys()
            yield '/', mount_dirs,[]
            for name in mount_dirs:
                for item in self.walk('/' + name, topdown, onerror):
                    yield item
        else:
            top_ospath = os.path.normpath(self.ospath(top))
            top_ospath_len = len(top_ospath)
            for dirpath, dirnames, filenames in os.walk(top_ospath,topdown,onerror):

                if self.dotfrs in dirnames:
                    dirnames.remove(self.dotfrs)

                if dirnames or filenames:
                    dir_sub_path = dirpath[top_ospath_len+1:].replace(os.path.sep,  '/')
                    if dir_sub_path:
                        yield self.joinpath(top, dirpath[top_ospath_len+1:].replace(os.path.sep,  '/')), dirnames, filenames
                    else:
                        yield top, dirnames, filenames

    # asset management

    def removeAsset(self, path):
        frsPath = self.frspath(path)
        if self.exists(frsPath):
            self.rmtree(frsPath) 

        if self.exists(path):
            if self.isfile(path):
                self.remove(path)
            else:
                self.rmtree(path)

        CacheMixin.removeCache(self, path)

    def moveAsset(self, src, dst):
        """ rename or move a file or folder
        """
        frsSrc = self.frspath(src)
        if self.exists(frsSrc):
            frsDst = self.frspath(dst)
            if not self.exists( self.dirname(frsDst) ):
                self.makedirs(self.dirname(frsDst))
            self.move(frsSrc, frsDst )

        if not self.exists( self.dirname(dst) ):
            self.makedirs( self.dirname(dst) )
        self.move(src, dst)

        CacheMixin.moveCache(self, src, dst)

    def copyAsset(self, src, dst, **kw):
        """ copy folder / file and associated subfiles, not include archives

        don't keep stat
        """
        if self.isfile(src):
            self.copyfile(src, dst)
        else:
            # copy folder
            if not self.exists(dst):
                self.makedirs(dst)
            for name in self.listdir(src):
                self.copyAsset(self.joinpath(src, name), self.joinpath(dst, name), copycache=0)

        # copy cache
        CacheMixin.copyCache(self, src,dst)

    def fullcopyAsset(self, src, dst):
        """ copy a folder or a file, include archives

        keep stat
        """
        if self.isfile(src):
            self.copyfile(src, dst)
        else:
            self.copytree(src, dst)

        frsSrc = self.frspath(src)
        if self.exists(frsSrc):
            frsDst = self.frspath(dst)
            frsDstDir = self.dirname(frsDst)
            if not self.exists(frsDstDir ):
                self.makedirs(frsDstDir)
            self.copytree(frsSrc, frsDst )




#coding:utf8
from config import FRS_REMOVED_FOLDER_NAME
from utils import timetag, int2ascii, ascii2int
import datetime
import time

class RecycleMixin:
    """ trushcase enables mixin for file repository system 
    """

    def removedpath(self, path, timestamp):
        if type(timestamp) in [str, unicode]:
            timestamp = ascii2int(timestamp)
        root, subpath= path[1:].split('/', 1)
        removeName = "%s.%s" % ( int2ascii(int(timestamp)), int2ascii(hash(subpath)))
        return self.joinpath('/',root, FRS_REMOVED_FOLDER_NAME, removeName)

    def recycleAssets(self, path, srcNames,timestamp=None):
        """ remove file or tree to trush folder """
        removedPath = self.removedpath(path, timestamp)
        for name in srcNames:
            srcPath = self.joinpath(path, name)
            dstPath = self.joinpath(removedPath, name)
            # remove cache file
            self.removeCache(srcPath)
            # just move to recycle
            self.moveAsset(srcPath, dstPath)

    def listRemoves(self, path):
        """ retime remved timestamps """
        recyle = self.joinpath(root, FRS_REMOVED_FOLDER_NAME)

        removes = []
        root, subpath= path[1:].split('/', 1)
        pathhash = int2ascii(hash(subpath))
        
        for filename in self.listdir(recyle):
            timestamp , _pathhash = filename.split('.', 1)
            if _pathhash == pathhash:
                removes.append(timestamp)
        return removes

    def listRemovedAssets(self, path, timestamp):
        removedPath = self.removedpath(path, timestamp)
        return self.listdir(removedPath)

    def revertRemove(self, path, removeName, assetName, reset_parent=None, new_name=''):
        """ 恢复一个删除的文件 """
        removedPath = self.removedpath(path, removeName)
        src_name = assetName 
        srcPath = self.joinpath(removedPath, src_name)
        dst_name = src_name if not new_name else new_name
        if reset_parent:
            dst_name = self.getNewName(reset_parent, dst_name)
            dstPath = self.joinpath(reset_parent, dst_name)
        else:
            dst_name = self.getNewName(path, dst_name)
            dstPath = self.joinpath(path, dst_name)

        self.moveAsset(srcPath, dstPath)

        # 已经全部恢复，删除这个文件夹
        if not self.listdir(removedPath):
            self.rmtree(removedPath)
        return dst_name

    def realRemove(self, path, removeName, assetNames=[]):
        removedPath = self.removedpath(path, removeName)
        if not self.exists(removedPath):
             return
        if not assetNames:
            self.rmtree(removedPath)
	    return 

        for name in assetNames:
            srcPath = self.joinpath(removedPath, name)
            self.removeAsset(srcPath)
        if not self.listdir(removedPath):
            self.rmtree(removedPath)

    def walkTrashBox(self, top_path='/', days=0, cmp_result=0):
        """ -1: older than history day; 0: all; 1: greater than history day 

        """
        # XXX need to rewrite
        cur_date = datetime.datetime(*time.gmtime()[:7])
        history_date = cur_date - datetime.timedelta(days)
        history_date_name = timetag(history_date)

        if top_path != '/':
            removeNames = self.listRemoves(top_path)
            if cmp_result != 0:
                removeNames = [removeName for removeName in removeNames
                            if cmp(removeName, history_date_name) == cmp_result]
            if removeNames:
                yield top_path, removeNames

        for parent_path, dirnames, filenames in self.walk(top_path):
            for dirname in dirnames:
                dir_vpath = self.joinpath(parent_path, dirname)
                removeNames = self.listRemoves(dir_vpath)
                if cmp_result != 0:
                    removeNames = [removeName for removeName in removeNames
                            if cmp(removeName, history_date_name) == cmp_result]
                if removeNames:
                    yield dir_vpath, removeNames

    def cleanTrushBox(self, top_path='/', days=30, cmp_result=-1):
        """ delete all items in trushboxes older than <keepDays> """
        for dir_vpath, remove_names in self.walkTrashBox(top_path, days, cmp_result):
            for remove_name in remove_names:
                self.realRemove(dir_vpath, remove_name)




# -*- encoding:utf-8 -*-
import time
import datetime
import os
import shutil
import socket
from random import random
from hashlib import md5
from config import FS_CHARSET
from types import UnicodeType
import array, binascii

def int2ascii(value):
    return binascii.b2a_hex(buffer(array.array('l', (value,))))  # 4byte

def ascii2int(value):
    bin = binascii.a2b_hex(value)
    index, sum = 0, 0
    for c in bin:
        sum += ord(c) << index
        index += 8
    return sum

def timetag(timestamp=None):
    if the_time is None:
        # use gmt time
        the_time = time.gmtime()
        return time.strftime('%Y-%m-%d-%H-%M-%S', the_time)
    else:
        the_time = datetime.datetime.fromtimestamp(timstamp)
        return the_time.strftime('%Y-%m-%d-%H-%M-%S')

def ucopy2(ossrc, osdst):
    # ucopy2 dosn't work with unicode filename yet
    if type(osdst) is UnicodeType and \
           not os.path.supports_unicode_filenames:
        ossrc = ossrc.encode(FS_CHARSET)
        osdst = osdst.encode(FS_CHARSET)
    shutil.copy2(ossrc, osdst)

def ucopytree(ossrc, osdst, symlinks=False):
    # ucopy2 dosn't work with unicode filename yet
    if type(osdst) is UnicodeType and \
            not os.path.supports_unicode_filenames:
        ossrc = ossrc.encode(FS_CHARSET)
        osdst = osdst.encode(FS_CHARSET)
    shutil.copytree(ossrc, osdst, symlinks)

def umove(ossrc, osdst):
    # umove dosn't work with unicode filename yet
    if type(osdst) is UnicodeType and \
           not os.path.supports_unicode_filenames:
        ossrc = ossrc.encode(FS_CHARSET)
        osdst = osdst.encode(FS_CHARSET)
    # windows move 同一个文件夹会有bug，这里改为rename
    # 例子： c:\test move to c:\Test
    if ossrc.lower() == osdst.lower():
        os.rename(ossrc, osdst)
    else:
        shutil.move(ossrc, osdst)

try:
    _v_network = str(socket.gethostbyname(socket.gethostname()))
except:
    _v_network = str(random() * 100000000000000000L)

def make_uuid(*args):
    t = str(time.time() * 1000L)
    r = str(random()*100000000000000000L)
    data = t +' '+ r +' '+ _v_network +' '+ str(args)
    uid = md5(data).hexdigest()
    return uid

def rst2frs(filename):
    """ 读取rst的参数，自动补充frs的json元数据 """

    # TODO

    # 读取大标题
    # 读取描述信息
    # 读取作者
    # 读取关键字
    # 当前时间为创建时间
    # 写入.frs文件夹




# for adapters
FRS_REMOVED_FOLDER_NAME = '.frsremoved'
FRS_ARCHIVED_FOLDER_NAME = 'archived'
FRS_CACHE_FOLDER_PREFIX = '.frs.'

import sys
# FS_CHARSET = 'gbk'
FS_CHARSET = sys.getfilesystemencoding()



################################
import os
import xml.sax.saxutils
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()

setup (
    name='zopen.frs',
    version='1.1',
    author = "panjy",
    author_email = "panjunyong@gmail.com",
    description = "zopen frs management",
    long_description=README,
    license = "Private",
    keywords = "zope3 z3c rpc server client operation",
    classifiers = [
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        ],
    url = 'http://pypi.zopen.cn/zopen.frs',
    packages = find_packages(),
    include_package_data = True,
    namespace_packages= ['zopen', ],
    install_requires = [
         'setuptools',
         # tests only
        'zope.datetime',
        ],
    zip_safe = False,
    entry_points = """
    """
)



# encoding: utf-8
import os
import shutil

from config import FRS_CACHE_FOLDER_PREFIX, FRS_ARCHIVED_FOLDER_NAME
from utils import ucopytree

class CacheMixin:

    def getCacheFolder(self, vpath, cachename=None):
        """ get os path of cache folder for vpath """
        cachebase = os.path.join(self.cache_root, 
                            *vpath.split('/') )

        if cachename:
            foldername = FRS_CACHE_FOLDER_PREFIX + cachename
            return os.path.join(cachebase, foldername)
        else:
            return cachebase

    def getCacheDecompressFileVpath(self, vpath, cachename=None, subpath=''):
        return '/cache' + vpath + '/' + FRS_CACHE_FOLDER_PREFIX + cachename + '/decompress' + subpath

    # TODO remove
    def getCacheDecompress(self, vpath, cachename=None):
        """ get os path of cache decompress for vpath """
        return os.path.join(self.getCacheFolder(vpath, cachename), 'decompress')

    # TODO remove
    def getCacheDecompressPreview(self, vpath, cachename=None):
        """ get os path of cache decompress preview for vpath """
        return os.path.join(self.getCacheFolder(vpath, cachename), 'decompresspreview')

    def hasCache(self, vpath, cachename=None):
        return os.path.exists(self.getCacheFolder(vpath, cachename))

    def removeCache(self, vpath,cachename=None):
        while True:
            vpath = '/cache' + vpath
            if self.exists(vpath):
                self.rmtree(vpath)
            else:
                break

    def moveCache(self, src, dst):
        while True:
            src, dst = '/cache' + src, '/cache' + dst
            if self.exists(src):
                self.move(src, dst)
            else:
                break

    def copyCache(self, src, dst, **kw):
        while True:
            src, dst = '/cache' + src, '/cache' + dst
            if self.exists(src):
                if self.exists(dst):  # 有时候事务不干净，这个会存在
                    self.rmtree(dst)
                self.copytree(src, dst)
            else:
                break
